import argparse
import datetime
import hashlib
import logging
import aiohttp
import asyncio
import urllib.parse
from functools import wraps
from typing import Any, Dict, Optional, Text, Union, Collection

from sqlalchemy.orm import Session

from rasa.cli import utils as rasa_cli_utils
from rasa.constants import GLOBAL_USER_CONFIG_PATH

from rasa.nlu.training_data import TrainingData as NluTrainingData
from rasa.utils import common as rasa_utils

from rasax.community import config, utils, constants
from rasax.community.constants import (
    CONFIG_METRICS_ENABLED,
    CONFIG_METRICS_USER,
    SEGMENT_WRITE_KEY,
    CONFIG_FILE_METRICS_KEY,
    CONFIG_METRICS_DATE,
    CONFIG_METRICS_WELCOME_SHOWN,
)

from rasax.community.version import __version__

logger = logging.getLogger(__name__)

METRICS_HTTP_TIMEOUT = 10  # Seconds

LOCAL_START_EVENT = "Local X Start"
SERVER_START_EVENT = "Server X Start"
METRICS_CONSENT_EVENT = "Metrics Consent"
MODEL_TRAINED_EVENT = "Model Trained"
MODEL_UPLOADED_EVENT = "Model Uploaded"
MODEL_PROMOTED_EVENT = "Model Promoted"
MESSAGE_RECEIVED_EVENT = "Message Received"
MESSAGE_ANNOTATED_EVENT = "Message Annotated"
MESSAGE_FLAGGED_EVENT = "Message Flagged"
STORY_CREATED_EVENT = "Story Created"

MESSAGE_ANNOTATED_CONVERSATIONS = "conversations"
MESSAGE_ANNOTATED_NEW_DATA = "annotate_new_data"
MESSAGE_ANNOTATED_INTERACTIVE_LEARNING = "interactive_learning"

STORY_CREATED_TTYB = "talk_to_your_bot"
STORY_CREATED_INTERACTIVE_LEARNING = "interactive_learning"
STORY_CREATED_STORIES = "stories"


def are_metrics_enabled() -> bool:
    """Did the user consent to the metrics collection."""
    return config.metrics_collection_config.get(CONFIG_METRICS_ENABLED, False)


def metrics_id() -> Optional[Text]:
    """Random unique consistent id."""
    return config.metrics_collection_config.get(CONFIG_METRICS_USER, None)


def ensure_metrics_enabled(f):
    """Function decorator for metrics functions that only runs the decorated
    function if metrics are enabled, and if the metrics user ID is set. All
    exceptions generated by the decorated functions are catched and logged."""

    @wraps(f)
    def decorated(*args, **kwargs):
        try:
            if are_metrics_enabled() and metrics_id():
                f(*args, **kwargs)
        except Exception as e:
            logger.debug("Skipping metrics collection: {}".format(e))

    return decorated


async def _track_async(
    distinct_id: Text,
    event_name: Text,
    properties: Optional[Dict[Text, Any]] = None,
    context: Optional[Dict[Text, Any]] = None,
) -> None:
    """Sends the contents of an event to the /track Segment endpoint.
    Documentation: https://segment.com/docs/sources/server/http/

    Do not call this function from outside metrics.py! This function does not
    check if metrics are enabled or not.

    Args:
        distinct_id: Unique metrics ID.
        event_name: Name of the event.
        properties: Values to send along the event.
        context: Context information about the event.
    """

    async with aiohttp.ClientSession() as session:
        headers = {
            "Authorization": "Basic {}".format(
                utils.encode_base64(SEGMENT_WRITE_KEY + ":")
            ),
            "Content-Type": "application/json",
        }

        payload = {
            "userId": distinct_id,
            "event": event_name,
            "properties": properties or {},
            "context": context or {},
        }

        if utils.in_continuous_integration():
            logger.info("Skip sending metrics: running in a CI context.")
            return

        async with session.post(
            "https://api.segment.io/v1/track",
            raise_for_status=True,
            headers=headers,
            json=payload,
            timeout=METRICS_HTTP_TIMEOUT,
        ) as resp:
            data = await resp.json()
            if not data.get("success"):
                raise Exception("Error sending metrics: {}".format(data))


def _track_internal(
    distinct_id: Text,
    event_name: Text,
    properties: Optional[Dict[Text, Any]] = None,
    context: Optional[Dict[Text, Any]] = None,
) -> None:
    """Tracks an event.

    If there's no event loop running on the current thread, `_track_internal`
    will block until the tracking HTTP request is sent. Otherwise, the task
    will be scheduled to run and the function will immediately return.

    Do not call this function from outside metrics.py! This function does not
    check if metrics are enabled or not.

    Args:
        distinct_id: Unique metrics ID.
        event_name: Name of the event.
        properties: Values to send along the event.
        context: Context information about the event.
    """

    coro = _track_async(distinct_id, event_name, properties, context)
    # get_event_loop() will create a new (unstarted) loop if the current
    # thread has no active one.
    loop = asyncio.get_event_loop()

    if loop.is_running():
        # We are being called after Sanic's loop has started.
        loop.create_task(coro)
    else:
        # We are being called before Sanic's loop has started, therefore run
        # the coroutune using run_until_complete() (this will start and stop
        # the event loop).
        loop.run_until_complete(coro)


@ensure_metrics_enabled
def track(
    event_name: Text,
    properties: Optional[Dict[Text, Any]] = None,
    context: Optional[Dict[Text, Any]] = None,
) -> None:
    """If enabled, track certain statistical metrics."""

    if properties:
        properties["metrics_id"] = metrics_id()

    _track_internal(metrics_id(), event_name, properties, context)


def track_metrics_consent(
    user_id: Text, allow_metrics: bool, auto_accept: bool = False
) -> None:
    """Track metric consent."""

    # Use _track_internal() instead of track() to bypass are_metrics_enabled() check
    try:
        _track_internal(
            user_id,
            METRICS_CONSENT_EVENT,
            {"allow_metrics_collection": allow_metrics, "auto_accept": auto_accept},
        )
    except Exception as e:
        logger.debug("Skipping metrics collection: {}".format(e))


@ensure_metrics_enabled
def track_local_start(
    session: Session,
    project_path: Text,
    domain: Dict,
    nlu_data: NluTrainingData,
    stories: Collection[Any],
) -> None:
    """Tracks an event when the local Rasa X edition is started."""

    project = hashlib.sha256(project_path.encode("utf-8")).hexdigest()
    start_event = _get_rasa_x_start_event(session, project, domain, nlu_data, stories)
    track(LOCAL_START_EVENT, start_event)


@ensure_metrics_enabled
def track_server_start(
    session: Session, domain: Dict, nlu_data: NluTrainingData, stories: Collection[Any]
) -> None:
    """Tracks an event when the dockerized Rasa X edition is started."""

    start_event = _get_rasa_x_start_event(
        session, config.project_name, domain, nlu_data, stories
    )
    track(SERVER_START_EVENT, start_event)


def _get_rasa_x_start_event(
    session: Session,
    project: Text,
    domain: Optional[Dict],
    nlu_data: NluTrainingData,
    stories: Collection[Any],
) -> Dict[Text, Union[int, Text]]:
    if domain is None:
        domain = {}

    from rasax.community.services.event_service import EventService
    from rasax.community.services.model_service import ModelService
    from rasax.community.services.data_service import DataService

    event_service = EventService(session)
    model_service = ModelService(config.rasa_model_dir, session)
    data_service = DataService(session)

    num_conversations = event_service.get_conversation_metadata_for_all_clients().count
    num_models = model_service.get_model_count()
    lookup_tables = data_service.get_lookup_tables(project, include_filenames=True)
    num_lookup_table_files = len({table["filename"] for table in lookup_tables})
    num_lookup_table_entries = sum(len(table["elements"]) for table in lookup_tables)
    synonyms = data_service.get_entity_synonyms(project)
    num_synonyms = sum(len(entry["synonyms"]) for entry in synonyms)
    num_regexes = data_service.get_regex_features(project).count

    return {
        "project": project,
        "rasa_x": __version__,
        "num_intent_examples": len(nlu_data.intent_examples),
        "num_entity_examples": len(nlu_data.entity_examples),
        "num_actions": len(domain.get("actions", [])),
        "num_templates": len(domain.get("templates", [])),
        "num_slots": len(domain.get("slots", [])),
        "num_forms": len(domain.get("forms", [])),
        "num_intents": len(domain.get("intents", [])),
        "num_entities": len(domain.get("entities", [])),
        "num_stories": len(stories),
        "num_conversations": num_conversations,
        "num_models": num_models,
        "num_lookup_table_files": num_lookup_table_files,
        "num_lookup_table_entries": num_lookup_table_entries,
        "num_synonyms": num_synonyms,
        "num_regexes": num_regexes,
    }


@ensure_metrics_enabled
def track_story_created(referrer: Optional[Text]) -> None:
    """Tracks an event when a new story is created."""

    if not referrer:
        return

    origin = None
    path = urllib.parse.urlparse(referrer).path

    if path.startswith("/talk/lenient"):
        origin = STORY_CREATED_TTYB
    elif path.startswith("/talk/strict"):
        origin = STORY_CREATED_INTERACTIVE_LEARNING
    elif path.startswith("/talk"):
        origin = STORY_CREATED_TTYB
    elif path.startswith("/stories"):
        origin = STORY_CREATED_STORIES

    if origin:
        track(STORY_CREATED_EVENT, {"story_created_from": origin})


@ensure_metrics_enabled
def track_message_annotated(origin: Text) -> None:
    """Tracks an event when a message is annotated."""

    track(MESSAGE_ANNOTATED_EVENT, {"message_annotated_from": origin})


@ensure_metrics_enabled
def track_message_annotated_from_referrer(referrer: Optional[Text] = None) -> None:
    """Tracks an event when a message is annotated, using a 'Referer' HTTP
    header value to determine the origin of the event."""

    if not referrer:
        return

    path = urllib.parse.urlparse(referrer).path

    if path.startswith("/conversations"):
        origin = MESSAGE_ANNOTATED_CONVERSATIONS
    elif path.startswith("/data"):
        origin = MESSAGE_ANNOTATED_NEW_DATA

    track_message_annotated(origin)


@ensure_metrics_enabled
def track_message_received(username: Text, channel: Optional[Text]) -> None:
    """Tracks an event when a message is received."""

    track(
        MESSAGE_RECEIVED_EVENT,
        {
            "username": hashlib.sha256(username.encode("utf-8")).hexdigest(),
            "channel": channel or constants.DEFAULT_CHANNEL_NAME,
        },
    )


def metrics_collection_config(args: argparse.Namespace) -> Dict[Text, Any]:
    """Make sure the user consents to any metrics collection."""

    import questionary

    try:
        return rasa_utils.read_global_config_value(
            CONFIG_FILE_METRICS_KEY, unavailable_ok=False
        )
    except ValueError:
        pass  # swallow the error and ask the user

    allow_metrics = (
        questionary.confirm(
            "Rasa will track a minimal amount of anonymized usage information "
            "(like how often the 'train' button is used) to help us improve Rasa X. "
            "None of your training data or conversations will ever be sent to Rasa. "
            "Are you OK with Rasa collecting anonymized usage data?"
        )
        .skip_if(args.no_prompt, default=True)
        .ask()
    )

    if not args.no_prompt:
        rasa_cli_utils.print_success(
            "Your decision has been stored into {}.".format(GLOBAL_USER_CONFIG_PATH)
        )
    else:
        rasa_cli_utils.print_info(
            "By adding the '--no_prompt' parameter you agreed to allow Rasa to track "
            "and send anonymized usage information."
        )

    settings = _metrics_settings(allow_metrics)

    rasa_utils.write_global_config_value(CONFIG_FILE_METRICS_KEY, settings)
    track_metrics_consent(settings[CONFIG_METRICS_USER], allow_metrics, args.no_prompt)
    return settings


def _metrics_settings(
    enable_metrics: bool,
    metrics_user_id: Optional[Text] = None,
    consent_date: Optional[datetime.datetime] = None,
) -> Dict[Text, Any]:
    import uuid

    return {
        CONFIG_METRICS_ENABLED: enable_metrics,
        CONFIG_METRICS_USER: metrics_user_id or uuid.uuid4().hex,
        CONFIG_METRICS_DATE: consent_date or datetime.datetime.now(),
        CONFIG_METRICS_WELCOME_SHOWN: False,
    }


def set_metrics_config_from_env() -> None:
    """Set the metrics configuration based on environment variables."""

    config.metrics_collection_config = _metrics_settings(
        config.metrics_consent_given, config.metrics_user_id
    )
